/*****************************************************************************/
/*
 *  exq_sym.c - symbol file parser
 *
 *  Parts of this file are
 *    Copyright (c) 2000-2010 Steven Levine and Associates, Inc.
 *    Copyright (c) 2010 Richard L Walsh
 *
*/
/*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define INCL_DOS
#define INCL_DOSERRORS
#include <os2.h>

#include "sym.h"

#include "exq.h"

/*****************************************************************************/

/** defines **/

#define SYMFILE_CNT 16
#define SYMBUF_SIZE 4096
#define SYMPTR_CNT  512

/** typedefs **/

typedef struct _SYMFILE
{
  HMODULE   hMod;
  ULONG     seg;
  ULONG     lru;
  FILE*     fSym;
  ULONG     cbFile;
  ULONG     maxSymLth;
  ULONG     cntSyms;
  ULONG     offsThisSeg;
  ULONG     offsSymPtr;
  USHORT    segFlags;
  USHORT    sfFlags;
  char*     pBuffer;
} SYMFILE;

typedef struct _SYMVAR
{
  ULONG     targetAddr;
  ULONG     lastAddr;
  ULONG     offsSymPtr;
  char *    pSymDef;
  char *    pText;
  char *    pSymPtrBuf;
} SYMVAR;

/*****************************************************************************/

/** static data */

static SYMVAR   symVar;
static SYMFILE  asf[SYMFILE_CNT];
static char     symBuffer[SYMBUF_SIZE];

/*****************************************************************************/

void    PrintSymbolFromSymFile(char * pszFile, HMODULE hMod,
                               ULONG ulObjNum, ULONG ulOffset);
BOOL    Sym32(SYMFILE* psf, SYMVAR* pvar);
BOOL    Sym16(SYMFILE* psf, SYMVAR* pvar);
SYMFILE* InitSymFile(ULONG ulSeg, HMODULE hMod, char* pszFile);
BOOL    OpenSymFile(SYMFILE* psf, char* pszFile);
BOOL    FindSegment(SYMFILE* psf);
ULONG   GetNextSegment(SYMFILE* psf, ULONG offsSeg, SEGDEF* psd);
void    CloseSymFiles(void);

/*****************************************************************************/
/*  Symbol File handler                                                      */
/*****************************************************************************/

/* 23 May 08 SHL Add oversized .sym file support
 * Symbol table format uses 16-bit offsets to point at the symbol
 * pointer table and to point at SymDef entries.  This breaks if there
 * are more than approx 64KiB of SymDefs.  To workaround this,
 * we assume that the symbol pointer table ends at the next SegDef
 * or at the end of file.  We then use the difference between the
 * calculated offset of the end of the pointer table and the calculated
 * offset to the next SegDef to calculate an adjusting offset.
 * This offset is applied to symbol pointer table lookups.
 * We also assume that the SymDef offsets in the symbol pointer table
 * are monotonic and detect 64KiB boundary crossings to calculate an
 * adjusting offset.  This offset is applied to SymDef entry indexing.
 */

void    PrintSymbolFromSymFile(char * pszFile, HMODULE hMod,
                               ULONG ulObjNum, ULONG ulOffset)
{
  ULONG     ctr;
  ULONG     size;
  ULONG     offset;
  ULONG     SymDef;
  ULONG     SymDefLo;
  ULONG     SymDefHi;
  ULONG     SymPtrAdjust;
  ULONG     SymPtrLast;
  USHORT*   pSymPtr;
  SYMFILE*  psf;
  SYMVAR*   pvar;

  TimedBeep();

  /* Get a ptr to a SYMFILE struct containing info needed to
   * step through the symbols for the requested segment.
   */
  psf = InitSymFile(++ulObjNum, hMod, pszFile);
  if (!psf) {
    fputc('\n', hTrap);
    return;
  }

  SymDefLo = 0xffffffff;
  SymDefHi = 0;
  SymPtrAdjust = 0;
  SymPtrLast = 0;

  /* Init a SYMVAR struct to (minimally) reduce stack usage. */
  pvar = &symVar;
  memset(pvar, 0, sizeof(SYMVAR));
  pvar->targetAddr = ulOffset;
  pvar->offsSymPtr = psf->offsSymPtr;
  pvar->pText = szBuffer;
  pvar->pSymPtrBuf = bigBuf;

  /* Put something in the buffer in case there's no symbol at offset 0 */
  strcpy(pvar->pText, "Start");

  /* Step through each entry in the symbol pointer table
   * until we reach or pass the requested offset.
   */
  for (ctr = 0; ctr < psf->cntSyms; ctr++, pSymPtr++) {

    TimedBeep();

    /* Cache a maximum of SYMPTR_CNT entries from the symbol ptr table.
     * This makes a huge improvement in performance by vastly reducing
     * the number of seeks required for large symfiles.
     */
    if (!(ctr & (SYMPTR_CNT - 1))) {
      if (fseek(psf->fSym, pvar->offsSymPtr, SEEK_SET)) {
        fprintf(hTrap, " Can not seek to SymDef offset at %x ", pvar->offsSymPtr);
        break;
      }

      size = psf->cntSyms - ctr;
      if (size > SYMPTR_CNT)
        size = SYMPTR_CNT;
      size *= sizeof(USHORT);

      if (fread(pvar->pSymPtrBuf, size, 1, psf->fSym) != 1) {
        fprintf(hTrap, " Can not read SymDef offsets at %x ", pvar->offsSymPtr);
        break;
      }

      pvar->offsSymPtr += size;
      pSymPtr = (USHORT*)pvar->pSymPtrBuf;
    }

    /* If symbol entries crossed a 64K boundary, inc the offset adjustment. */
    if (*pSymPtr < SymPtrLast)
      SymPtrAdjust += 0x10000;
    SymPtrLast = *pSymPtr;

    /* Calc the offset of the next SymDef (relative to the seg's start). */
    SymDef = *pSymPtr + SymPtrAdjust;

    /* If the desired offset isn't within the SymDef cache, seek to that
     * offset in the file and refill the cache.  To avoid having to check
     * each SymDef16/32 to ensure it lies entirely within the cache, set
     * SymDefHi to an offset that could contain the largest possible SymDef.
     * Like the SymPtr cache, this cache produces a huge performance gain.
     */
    if (SymDef < SymDefLo || SymDef > SymDefHi) {
      SymDefLo = SymDef;

      /* Convert from a seg-relative to a file-relative offset. */
      offset = SymDefLo + psf->offsThisSeg;
      if (fseek(psf->fSym, offset, SEEK_SET)) {
        fprintf(hTrap, " Can not seek to SymDef at %lx ", offset);
        break;
      }

      /* Read the lesser of SYMBUF_SIZE bytes or the remaining SymDefs. */
      size = psf->offsSymPtr - SymDefLo;
      if (size > SYMBUF_SIZE) {
        size = SYMBUF_SIZE;
        SymDefHi = SymDefLo + size - (sizeof(SYMDEF32) + psf->maxSymLth);
      }
      else
        SymDefHi = SymDefLo + size - sizeof(SYMDEF32);

      if (fread(psf->pBuffer, 1, size, psf->fSym) < size) {
        fprintf(hTrap, " Can not read SymDef16/32 - %lx bytes at %lx ",
                size, offset);
        break;
      }
    }

    /* Calc the address within the cache where the desired SymDef lies. */
    pvar->pSymDef = &psf->pBuffer[SymDef - SymDefLo];

    /* Call the function that will determine if we've reached or exceeded
     * the desired offset.  If so, it will print out the data & return TRUE.
     * If not, it will store some info and return FALSE.
     */
    if (psf->segFlags & 0x01) {
      if (Sym32(psf, pvar))
        break;
    }
    else {
      if (Sym16(psf, pvar))
        break;
    }

    /* Special case for the last symbol in the table. */
    if (ctr + 1 == psf->cntSyms) {
      if (pvar->targetAddr >= pvar->lastAddr && *pvar->pText) {
        if (pvar->targetAddr == pvar->lastAddr)
          fprintf(hTrap, " %s (%X)", pvar->pText, pvar->targetAddr);
        else
          fprintf(hTrap, " near %s + %X ", pvar->pText, pvar->targetAddr - pvar->lastAddr);
      }
    }
  } /* for symbols */

  fputc('\n', hTrap);
  return;
}

/*****************************************************************************/

BOOL    Sym32(SYMFILE* psf, SYMVAR* pvar)
{
  ULONG     ul;
  SYMDEF32* psd32;

  psd32 = (SYMDEF32*)pvar->pSymDef;

  /* If the requested offset is GTE the previous symbol offset
   * and LT the current symbol offset, then it's time to display.
   */
  ul = pvar->targetAddr >= pvar->lastAddr && pvar->targetAddr < psd32->wSymVal;
  if (ul) {

    if (pvar->targetAddr == pvar->lastAddr) {
      fprintf(hTrap, " %s", pvar->pText);
      return TRUE;
    }

    fprintf(hTrap, " between %s + %X ",
            pvar->pText, pvar->targetAddr - pvar->lastAddr);
  }

  /* Remember symbol and offset */
  pvar->lastAddr = psd32->wSymVal;

  memcpy(pvar->pText, psd32->achSymName, psd32->cbSymName);
  pvar->pText[psd32->cbSymName] = 0;

  /* Display if requested offset is between symbols or at symbol. */
  if (ul) {
    fprintf(hTrap, "and %s - %X",
            pvar->pText, pvar->lastAddr - pvar->targetAddr);
    return TRUE;
  }

  return FALSE;
}

/*****************************************************************************/

BOOL    Sym16(SYMFILE* psf, SYMVAR* pvar)
{
  ULONG     ul;
  SYMDEF16* psd16;

  psd16 = (SYMDEF16*)pvar->pSymDef;

  /* If the requested offset is GTE the previous symbol offset
   * and LT the current symbol offset, then it's time to display.
   */
  ul = pvar->targetAddr >= pvar->lastAddr && pvar->targetAddr < psd16->wSymVal;
  if (ul) {

    if (pvar->targetAddr == pvar->lastAddr) {
      fprintf(hTrap, " %s", pvar->pText);
      return TRUE;
    }

    fprintf(hTrap, " between %s + %X ", pvar->pText, pvar->targetAddr - pvar->lastAddr);
  }

  /* Remember symbol and offset */
  pvar->lastAddr = psd16->wSymVal;

  memcpy(pvar->pText, psd16->achSymName, psd16->cbSymName);
  pvar->pText[psd16->cbSymName] = 0;

  /* Display if requested offset is between symbols or at symbol. */
  if (ul) {
    fprintf(hTrap, "and %s - %X", pvar->pText, pvar->lastAddr - pvar->targetAddr);
    return TRUE;
  }

  return FALSE;
}

/*****************************************************************************/

SYMFILE* InitSymFile(ULONG ulSeg, HMODULE hMod, char* pszFile)
{
  static ULONG  ulLRU = 0;

  BOOL      ok;
  SYMFILE*  psf;
  SYMFILE*  psfSave;
  SYMFILE*  psfLRU;

  if (!hMod)
    return 0;

  ulLRU++;
  psfSave = 0;
  psfLRU = 0;

  /* Search the list of open symfiles for a match.  If one points to the
   * correct file & seg, reuse it.  If the file matches but not the seg,
   * save it to reuse the open file.  Exit at the first unused entry.
   */
  for (psf = asf; psf < &asf[SYMFILE_CNT] && psf->hMod; psf++) {
    if (psf->hMod == hMod) {
      if (psf->seg == ulSeg) {
        psf->lru = ulLRU;
        return psf;
      }
      psfSave = psf;
    }

    if (!psfLRU || psf->lru < psfLRU->lru)
      psfLRU = psf;
  }

  /* If all of the slots are in use, recycle the least recently used entry. */
  if (psf >= &asf[SYMFILE_CNT]) {
    if (psfLRU != psfSave) {
      if (!psfLRU->sfFlags && psfLRU->fSym)
        fclose(psfLRU->fSym);
      memset(psfLRU, 0, sizeof(SYMFILE));
    }
    psf = psfLRU;
  }

  psf->hMod = hMod;
  psf->seg  = ulSeg;
  psf->lru = ulLRU;
  psf->pBuffer = symBuffer;

  /* If we're reusing an existing file, copy some data and flag both the
   * original & duplicate entries to prevent the file from being closed.
   * Otherwise, open the new file.
   */
  if (psfSave) {
    psfSave->sfFlags = 1;
    psf->fSym = psfSave->fSym;
    psf->cbFile = psfSave->cbFile;
    psf->sfFlags = 1;
    ok = TRUE;
  }
  else
    ok = OpenSymFile(psf, pszFile);

  /* Find the requested seg */
  if (ok)
    ok = FindSegment(psf);

  /* If there was an error, close any newly-opened file and clear the entry */
  if (!ok) {
    if (!psf->sfFlags && psf->fSym)
      fclose(psf->fSym);
    memset(psf, 0, sizeof(SYMFILE));
    psf = 0;
  }

  return psf;
}

/*****************************************************************************/

BOOL    OpenSymFile(SYMFILE* psf, char* pszFile)
{
  char *    ptr;

  strcpy(psf->pBuffer, pszFile);

  /* Replace the file extension with '.sym' */
  ptr = strrchr(psf->pBuffer, '.');
  if (!ptr)
    ptr = strchr(psf->pBuffer, 0);
  strcpy(ptr, ".SYM");

  /* If it can't be opened using the f/q path, try the current directory. */
  psf->fSym = fopen(psf->pBuffer, "rb");
  if (!psf->fSym) {
    ptr = strrchr(psf->pBuffer, '\\');
    if (ptr)
      psf->fSym = fopen(ptr + 1, "rb");
  }

  if (!psf->fSym)
    return FALSE;

  /* Save the file's size.  FindSegment() will do a SEEK_SET. */
  fseek(psf->fSym, 0, SEEK_END);
  psf->cbFile = (ULONG)ftell(psf->fSym);
  if (psf->cbFile == (ULONG)-1) {
    fprintf(hTrap, " Can not determine .sym file size ");
    return FALSE;
  }

  return TRUE;
}

/*****************************************************************************/

BOOL    FindSegment(SYMFILE* psf)
{
  BOOL      ok;
  ULONG     offsNextSeg;
  ULONG     pSymDef;
  ULONG     offs;
  ULONG     adjust;
  MAPDEF *  pmd;
  SEGDEF *  psd;

  fseek(psf->fSym, 0, SEEK_SET);

  /* Read the MAPDEF at the beginning of the file. */
  pmd = (MAPDEF*)psf->pBuffer;
  if (fread(pmd, sizeof(MAPDEF), 1, psf->fSym) != 1) {
    fprintf(hTrap, " Can not read MAPDEF ");
    return FALSE;
  }

  if (psf->seg > pmd->cSegs) {
    fprintf(hTrap, " Requested segment nbr %ld is out of range ", psf->seg);
    return FALSE;
  }

  psf->offsThisSeg = pmd->ppSegDef * 16;
  psf->maxSymLth = pmd->cbMaxSym;

  /* Step through each SEGDEF until we find the one we need,
   * or we find it isn't present, or we reach the end of the file.
   */
  ok = FALSE;
  psd = (SEGDEF*)psf->pBuffer;
  while (psf->offsThisSeg && psf->offsThisSeg < psf->cbFile) {

    psf->offsThisSeg = GetNextSegment(psf, psf->offsThisSeg, psd);
    if (!psf->offsThisSeg)
      break;

    if (psd->wSegNum == psf->seg) {
      ok = TRUE;
      break;
    }

    if (psd->wSegNum > psf->seg) {
      fprintf(hTrap, " Can not find segment nbr %ld ", psf->seg);
      break;
    }

    psf->offsThisSeg = psd->ppNextSeg * 16;
  }

  if (!ok)
    return FALSE;

  /* Save some info we'll need, then find the offset of the next seg. */
  psf->cntSyms  = psd->cSymbols;
  psf->segFlags = psd->bFlags;
  pSymDef = psd->pSymDef;

  if (!psd->ppNextSeg)
    offsNextSeg = psf->cbFile - sizeof(LAST_MAPDEF);
  else
    offsNextSeg = GetNextSegment(psf, psd->ppNextSeg * 16, psd);

  if (!offsNextSeg)
    return FALSE;

#define XQ_OVERHEAD (sizeof(SYMDEF32) - 1 + (2 * sizeof(USHORT)))

  /* See if there's any chance that cntSym has rolled over because there are
   * over 64k symbols.  If so, the average length of a symbol name will be
   * absurdly high.  Test the length against the lesser of MAPDEF->cbMaxSym
   * and 80.  Note: 80 may be too high but heavily mangled C++ symbols that
   * use multiple namespaces could drive it close to that length. RLW
  */
  if (offsNextSeg - psf->offsThisSeg > 0x10000 * (XQ_OVERHEAD + 3)) {
    adjust = (psf->maxSymLth > 80 ? 80 : psf->maxSymLth) + XQ_OVERHEAD;
    while (((offsNextSeg - psf->offsThisSeg) / psf->cntSyms) > adjust)
      psf->cntSyms += 0x10000;
  }

  /* The end of the symbol pointer table should be roughly equal to the
   * offset of the next SEGDEG, but it will be off by some multiple of
   * 64k if this seg has more than 64k of symbols.  This calculates the
   * adjustment needed to correctly locate the start of the table,  RLW
   */
  offs = psf->offsThisSeg + pSymDef + (2 * psf->cntSyms * sizeof(USHORT));
  adjust = (offsNextSeg - offs) & 0xFFFF0000;
  psf->offsSymPtr = psf->offsThisSeg + pSymDef + adjust;

  return TRUE;
}

/*****************************************************************************/

ULONG   GetNextSegment(SYMFILE* psf, ULONG offsSeg, SEGDEF* psd)
{
  /* Look for the next SegDef starting at the offset given, then validate it.
   * If validation fails, it may be because the current seg has more than 1mb
   * of symbol data (the maximum this file format was designed to handle). If
   * so, keep seeking 1mb ahead until validation succeeds or we reach EOF. RLW
   */
  while (offsSeg < psf->cbFile) {

    if (fseek(psf->fSym, offsSeg, SEEK_SET)) {
      fprintf(hTrap, " Can not seek to SegDef at %x ", offsSeg);
      return 0;
    }

    if (fread(psd, sizeof(SEGDEF), 1, psf->fSym) != 1) {
      fprintf(hTrap, " Can not read SegDef at %x ", offsSeg);
      return 0;
    }

    if (!psd->wInstance0 && !psd->wInstance1 && !psd->wInstance2 &&
        !psd->bReserved1 && psd->cSymbols && psd->pSymDef &&
        psd->wSegNum && psd->cbSegName && psd->achSegName[0]) {
      return offsSeg;
    }

    offsSeg += 0x100000;
  }

  fprintf(hTrap, " Reached end-of-file looking for segment nbr %ld ",
          psf->seg);

  return 0;
}

/*****************************************************************************/

void    CloseSymFiles(void)
{
  SYMFILE*  psf;
  SYMFILE*  psf2;

  for (psf = asf; psf < &asf[SYMFILE_CNT]; psf++) {
    if (!psf->fSym)
      continue;

    fclose(psf->fSym);

    /* ensure a file with multiple entries only gets closed once */
    for (psf2 = psf + 1; psf2 < &asf[SYMFILE_CNT]; psf2++) {
      if (psf2->fSym == psf->fSym)
        psf2->fSym = 0;
    }
  }

  return;
}

/*****************************************************************************/

